import sqlite3

auth_queries = {
    '+user': 'INSERT OR REPLACE INTO users (uniqueid, implicit_level) VALUES (?, ?)',
    '-user': 'DELETE FROM users WHERE uniqueid=?',
    '+group': 'INSERT OR REPLACE INTO groups (name, implicit_level) VALUES (?, ?)',
    '-group': 'DELETE FROM groups WHERE name=?',
    '+perm': 'INSERT OR REPLACE INTO permissions (name, implicit_level) VALUES (?,?)',
    '-perm': 'DELETE FROM permissions WHERE name=?',

    '+member': 'INSERT OR IGNORE INTO group_members VALUES ((SELECT id FROM users WHERE uniqueid=?), (SELECT id FROM groups WHERE name=?))',
    '-member': 'DELETE FROM group_members WHERE user_id=(SELECT id FROM users WHERE uniqueid=?) AND group_id=(SELECT id FROM groups WHERE name=?)',

    '+group_perm': 'INSERT OR REPLACE INTO group_permissions VALUES ((SELECT id FROM groups WHERE name=?), (SELECT id FROM permissions WHERE name=?))',
    '-group_perm': 'DELETE FROM group_permissions WHERE group_id=(SELECT id FROM groups WHERE name=?) AND permission_id=(SELECT id FROM permissions WHERE name=?)',

    '+user_perm': 'INSERT OR REPLACE INTO user_permissions VALUES ((SELECT id FROM users WHERE uniqueid=?), (SELECT id FROM permissions WHERE name=?), ?)',
    '-user_perm': 'DELETE FROM user_permissions WHERE user_id=(SELECT id FROM users WHERE uniqueid=?) AND permission_id=(SELECT id FROM permissions WHERE name=?))',
    
    'list_users': 'SELECT uniqueid, implicit_level FROM users',
    'list_perms': 'SELECT name, implicit_level FROM permissions',

    '?user': 'SELECT id, implicit_level FROM users WHERE uniqueid=? LIMIT 1',
    '?perm': 'SELECT id, implicit_level FROM permissions WHERE name=? LIMIT 1',
    '?group': 'SELECT id, implicit_level FROM groups WHERE name=? LIMIT 1',

    '?user_perm': 'SELECT permitted FROM user_permissions WHERE user_id=? AND permission_id=? LIMIT 1',
    '?user_group_perms': 'SELECT group_id FROM group_permissions WHERE permission_id=? AND group_id IN (SELECT group_id FROM group_members WHERE user_id=?) LIMIT 1',
    '?group_level': 'SELECT max(implicit_level) FROM groups WHERE id IN (SELECT group_id FROM group_members WHERE user_id=?)',
}

class AuthDB(object):
    def __init__(self):
        self.connection = sqlite3.connect(':memory:')
        self.cursor = self.connection.cursor()
        self.setup()

    def setup(self):
        with open('auth.sql', 'r') as f:
            self.cursor.executescript(f.read())

    def check(self, name, perm):
        perm = self.execute('?perm', perm).fetchone()
        if not perm:
            # Permission does not exist - deny
            return False
        pid, perm_level = perm
        user = self.execute('?user', name).fetchone()
        if not user:
            # User has no specific permission, so allow if public
            return perm_level == 0
        uid, user_level = user

        # Check user specific permissions
        user_permitted = self.execute('?user_perm', uid, pid).fetchone()
        if user_permitted:
            return user_permitted[0]

        # Check if any of the users groups have this permission
        group_permitted = self.execute('?user_group_perms', pid, uid).fetchone()
        if group_permitted:
            return True
        group_level = self.execute('?group_level', uid).fetchone()
        
        return (group_level and group_level[0] >= perm_level) or user_level >= perm_level

    def execute(self, key, *args):
        return self.cursor.execute(auth_queries[key], args)

    def add_user_permission(self, user, perm, permitted=False):
        self.execute('+user_perm', user, perm, permitted)

    def remove_user_permission(self, user, perm,):
        self.execute('-user_perm', user, perm)
        
    def add_group_permission(self, group, perm):
        self.execute('+group_perm', group, perm)

    def remove_group_permission(self, group, perm):
        self.execute('-group_perm', group, perm)

    def add_user_to_group(self, uid, group_name):
        self.execute('+member', uid, group_name)

    def remove_user_from_group(self, uid, group_name):
        self.execute('-member', uid, group_name)
        
    def add_permission(self, name, implicit_level=None):
        self.execute('+perm', name, implicit_level)

    def remove_permission(self, name):
        self.execute('DELETE FROM permissions WHERE name=?', (name,))
        
    def add_user(self, uid, implicit_level=None):
        self.execute('+user', uid, implicit_level)

    def remove_user(self, uid):
        self.execute('-user', uid)

    def add_group(self, name, implicit_level=None):
        self.execute('+group', name, implicit_level)
        
    def remove_group(self, name):
        self.execute('-group', name)

        
auth = AuthDB()

if __name__ == '__main__':
    # Adding a user
    auth.add_user('DanielB', 50)
    # Adding a group
    auth.add_group('Scripters')
    
    # Adding a user to a group
    auth.add_user_to_group('DanielB', 'Scripters')
    
    # Adding permissions
    auth.add_permission('always_allow', 0)
    auth.add_permission('high_level_required', 100)
    auth.add_permission('disallow_on_user', 2)
    auth.add_permission('allow_on_user', 999)
    
    # No explicit permission, but my implicit level is higher
    assert auth.check('DanielB', 'always_allow')
    # No explicit permission, but my implicit level is lower
    assert not auth.check('DanielB', 'high_level_required')
    
    # I should be able to use this, because my level is higher
    assert auth.check('DanielB', 'disallow_on_user')
    # But now I shouldn't be able
    auth.add_user_permission('DanielB', 'disallow_on_user', False)
    assert not auth.check('DanielB', 'disallow_on_user')
    
    # This I shouldn't be able
    assert not auth.check('DanielB', 'allow_on_user')
    # Until I allow it
    auth.add_user_permission('DanielB', 'allow_on_user', True)
    assert auth.check('DanielB', 'allow_on_user')

    
